// $Id: Sf2Smt.cpp,v 1.9 2008/11/17 21:29:09 root Exp $
//
// Copyright (C) 2007. Domagoj Babic, see the LICENSE.

#include "Sf2Smt.h"
#include <cassert>
#include <cstring>
#include <sstream>
#include <fstream>
#include <iterator>
#include <algorithm>

using namespace std;

const static struct keywords {
    char *Sf;
    char *SmtPred;
    char *SmtBv;
    bool predicate;
} keywords[] = {
	{"%s",      "",             "bvsrem",       false},
	{"%u",      "",             "bvurem",       false},
	{"&",       "and",          "bvand",        false},
	{"*",       "",             "bvmul",        false},
	{"+",       "",             "bvadd",        false},
	{"-",       "",             "bvsub",        false},
	{"/=",      "xor",          "distinct",     true},
	{"/s",      "",             "bvsdiv",       false},
	{"/u",      "",             "bvudiv",       false},
	{"<<",      "",             "bvshl",        false},
	{"=",       "iff",          "=",            true},
	{"=>",      "implies",      "",             true},
	{">>a",     "",             "bvashr",       false},
	{">>l",     "",             "bvlshr",       false},
	{"^",       "xor",          "bvxor",        false},
	{"conc",    "",             "concat",       false},
	{"extr",    "",             "extract",      false},
	{"ite",     "if_then_else", "ite",          false},
	{"sext",    "",             "sign_extend",  false},
	{"sge",     "",             "bvsge",        true},
	{"sgt",     "",             "bvsgt",        true},
	{"sle",     "",             "bvsle",        true},
	{"slt",     "",             "bvslt",        true},
	{"trun",    "",             "",             false}, // Synth!
	{"uge",     "",             "bvuge",        true},
	{"ugt",     "",             "bvugt",        true},
	{"ule",     "",             "bvule",        true},
	{"ult",     "",             "bvult",        true},
	{"zext",    "",             "zero_extend",  false},
	{"|",       "or",           "bvor",         false},
	{"~",       "not",          "bvnot",        false}
};

static const struct keywords *lookup(const char *txt) {
    int low = 0;
    int mid;
    int high = sizeof(keywords) / sizeof(keywords[0]) - 1;
    while (low <= high) {
        int c;
        mid = low + (high - low) / 2;
        if ((c = strcasecmp(keywords[mid].Sf, txt)) == 0) {
            return &keywords[mid];
        } else if (c < 0) {
            low = mid + 1;
        } else {
            high = mid - 1;
        }    
    }
    return NULL;
}

static void eatLine(istream &IS) {
    while (!IS.eof() && IS.get() != '\n');
}

static void eatSpace(istream &IS) {
    while (!IS.eof() && (IS.peek() == '\t' || IS.peek() == ' ')) {
        (void)IS.get();
    }
}

string Parser::parseVarDef(istream &IS) {
    eatSpace(IS);
    string Name;
    while (!IS.eof() && IS.peek() != ':' && IS.peek() != '\n' &&
            IS.peek() != ' ') {
        Name.push_back(IS.get());
    }
    assert(!Name.empty() && "Nameless variables not allowed.");
    return Name;
}

void Parser::parseConstant(istream &IS, uint64_t &val) {
    eatSpace(IS);
    uint64_t x = 0;
    while (IS.peek() >= '0' && IS.peek() <= '9') {
        x = 10 * x + (IS.get() - '0');
    }
    val = x;
}

void Parser::parseOperands(istream &IS, vector<pair<string, unsigned> >
        &Ops, bool EncodeConstantsDirectly) {
    eatSpace(IS);
    while (!IS.eof() && IS.peek() != '\n') {
        unsigned bits = 0;
        string Op;
        if (IS.peek() >= '0' && IS.peek() <= '9') { // Constant
            stringstream tmp;
            uint64_t val = 0;
            parseConstant(IS, val);
            (void)IS.get(); // Eat ':'
            (void)IS.get(); // Eat 'i'
            IS >> bits; // Parse number of bits
            // For extr 2nd and 3rd params, encode only constants.
            if (EncodeConstantsDirectly && !Ops.empty()) {
                if (Ops.size() == 1) { // First operand
                    tmp << val;
                } else { // Second operand
                    tmp << (val - 1);
                } 
            } else {
                tmp << "bv" << val << '[' << bits << "]";
            }
            Op = tmp.str();
        } else {
            IS >> Op;
            iterator I = Symbols.find(Op);
            if (I == Symbols.end()) {
                cerr << "Undefined symbol: " << Op << "\n";
                exit(EXIT_FAILURE);
            }
            bits = I->second;
        }
        Ops.push_back(make_pair(Op, bits));
        eatSpace(IS);
    }
}

void Parser::parseHeader(istream &IS) {
    while (true) {
        if (IS.peek() == '#') { eatLine(IS); }
        else if (IS.peek() == 'v') { // Version Line
            (void)IS.get(); // Skip 'v'
            // Check conformance
            float Version = 0.0;
            IS >> Version;
            assert(Version == 1.0 && 
                "Invalid version, Sf2Smt requires an update.");
            eatLine(IS); // Eat the rest of the line
        } else if (IS.peek() == 'e') { // Expectation line
            // SMT has no notion of expectation - just discard the line.
            eatLine(IS);
        } else if (IS.peek() == 'd') { // Definition line
            (void)IS.get(); // Skip 'd'
            while (IS.peek() != '\n') {
                // All declared variables, even boolean ones, are
                // declared as bit-vectors.
                const string Name = parseVarDef(IS);
                assert(!Name.empty() && "Nameless variables not allowed.");
                char dummy;
                IS >> dummy;    // Eat ':'
                IS >> dummy;    // Eat 'i'
                unsigned Bits;
                IS >> Bits;
                eatSpace(IS);
                
                stringstream ss;
                ss << ":extrafuns ((" << Name << " BitVec[" << Bits <<
                    "]))\n";
                Header.push_back(ss.str());
                Symbols.insert(make_pair(Name, Bits));
            }
            eatLine(IS); // Eat '\n'
        } else { // Probably a constraint or a predicate
            assert((IS.peek() == 'c' || IS.peek() == 'p')
            && "Found line that's neither predicate nor constraint.");
            return;
        }
    }
}

namespace {

// Select 1st from the pair.
template <typename Pair>
struct select1st {
    typedef typename Pair::first_type result_type;
    const result_type &operator()(const Pair &p) const {
        return p.first;
    }
};

} // End of anonymous namespace

string Parser::encodeOperator(string &Operator, 
        vector<pair<string, unsigned> > &Ops,
        unsigned ResultBits) {

    const struct keywords *kw = lookup(Operator.c_str());
    if (!kw) {
        cerr << "Operator: " << Operator << " not found.\n";
        exit(EXIT_FAILURE);
    }
    stringstream ss;

    if (Operator.compare("trun") == 0) {
        assert(Ops.size() == 1 && "Invalid number of operators.");
        ss << '(' << kw->SmtBv << "extract[" << (ResultBits-1) << ":0] " << 
            Ops[0].first << ')';
    } else if (Operator.compare("sext") == 0 || 
               Operator.compare("zext") == 0) {
        assert(Ops.size() == 1 && "Invalid number of operators.");
        unsigned bits = Ops[0].second;
        ss << '(' << kw->SmtBv << "[" << (ResultBits - bits) 
            << "] " << Ops[0].first << ')';
    } else if (Operator.compare("ite") == 0) {
        // First operand must be boolean

        assert(Ops.size() == 3 && "Invalid number of operators.");
        ss << "(ite (= " << Ops[0].first << " bv1[1]) " << 
            Ops[1].first << ' ' << Ops[2].first << ')';
    } else if (Operator.compare("=>") == 0) {
        // Both operands must be boolean

        assert(Ops.size() == 2 && "Invalid number of operators.");
        ss << '(' << kw->SmtPred << " (= " << Ops[0].first << 
            " bv1[1]) (= " << Ops[1].first << " bv1[1]))";
    } else if (Operator.compare("extr") == 0) {
        assert(Ops.size() == 3 && "Invalid number of operators.");
        ss << "(extract[" << Ops[2].first << ':' << Ops[1].first 
            << "] " << Ops[0].first << ')';
    } else {
        ss << '(' << kw->SmtBv << ' ';
        transform(Ops.begin(), Ops.end(),
                ostream_iterator<string>(ss, " "),
                select1st<pair<string,unsigned> >());
        ss << ')';
    }

    if (kw->predicate) {
        // Result must be boolean

        // ss contains the operator encoding.
                
        // Create a fresh var
        stringstream fresh, line;

        do {
            fresh.str("");
            fresh << "Fresh__" << Cnt++;
        } while (Symbols.find(fresh.str()) != Symbols.end());
        Symbols.insert(make_pair(fresh.str(), 1));

        line << ":extrafuns ((" << fresh.str() << " BitVec[1]))\n";
        // Create definition of the new var
        Header.push_back(line.str());

        // Clean up stringstream
        line.str("");
        // Create constraint
        line << ":assumption (iff (= " << fresh.str() << " bv1[1]) " <<
            ss.str() << ")\n";
        Body.push_back(line.str());
        ss.str("");
        ss << fresh.str();
    }

    return ss.str();
}

void Parser::parseLine(istream &IS) {
    char dummy = 'x';
    IS >> dummy;
    if (dummy == '#') { 
        eatLine(IS); 
        return; 
    }
    if (dummy == 'c') { // Constraint
        unsigned bits = 0;
        string Res, Operator;
        eatSpace(IS);

        if (IS.peek() >= '0' && IS.peek() <= '9') {
            // Result is a constant
            stringstream tmp;
            uint64_t val = 0;
            parseConstant(IS, val);
            (void)IS.get(); // Eat ':'
            (void)IS.get(); // Eat 'i'
            IS >> bits; // Parse number of bits
            // Added by E. Pavlenko as a junk free line
            tmp << "bv" << val << '[' << bits << "]";     
            Res = tmp.str();
        } else {
            // Result is a variable
            IS >> Res;
            iterator I = Symbols.find(Res);
            assert(I != Symbols.end() && "Undefined symbol.");
            bits = I->second;
        }

        IS >> Operator;
        vector<pair<string, unsigned> > Ops; 
        // For extraction, operands must be simple integers.
        parseOperands(IS, Ops, Operator.compare("extr") == 0);
        eatLine(IS); // Eat '\n'

        stringstream ss;
        ss << ":assumption (= " << Res << ' ' << 
            encodeOperator(Operator, Ops, bits) << ")\n";
        Body.push_back(ss.str());

    } else if (dummy == 'p') { // Predicate
        string Operator;
        IS >> Operator;
        vector<pair<string, unsigned> > Ops; 
        parseOperands(IS, Ops);
        eatLine(IS); // Eat '\n'

        stringstream ss;
        ss << ":assumption (= bv1[1] " << 
            encodeOperator(Operator, Ops, 1) << ")\n";
        Body.push_back(ss.str());

    } else if (!IS.eof()) {
        assert((dummy == '\t' || dummy == ' ')
        && "Only empty lines, or lines starting with v,e,d,p,c,# allowed.");
        eatSpace(IS);
        assert(IS.get() == '\n'
        && "Only empty lines, or lines starting with v,e,d,p,c,# allowed.");
    }
}

void Parser::parse(std::istream &IS) {
    parseHeader(IS);
    while (!IS.eof()) {
        parseLine(IS);
    }
}

ostream &operator<<(ostream &OS, const Parser &P) {
    OS << "( benchmark " << P.Name << '\n';
    OS << ":source { Spear Sf2Smt translator }\n";
    OS << ":logic QF_BV\n";
    copy(P.Header.begin(), P.Header.end(),
        ostream_iterator<string>(OS, ""));
    copy(P.Body.begin(), P.Body.end(),
        ostream_iterator<string>(OS, ""));
    OS << ")\n";
    return OS;
}






// ---------- MAIN ------------

int main(int argc, char **argv) {
    if (argc > 1) {

        if (!strcmp("--stdin", argv[1])) {
            Parser P("stdin");
            P.parse(std::cin);
            cout << P;
        } else {
            Parser P(basename(argv[1]));
            ifstream in(argv[1]);
            if (!in) {
                cerr << "Error: Could not open smt file: " << argv[1] << endl;
                return EXIT_FAILURE;
            }
            P.parse(in);
            cout << P;
        }
    } else {
        std::cout << "Usage: sf2smt [--stdin | input_file]" << std::endl;
        return EXIT_SUCCESS;
    }
}
